/*
 * Freefall.h - Includes all functions required for programming the Freefall board.
 *
 *   Copyright (c) Stingworks 2010
 *   License: MIT License 
 *
 * File History:
 *   21	 JULY  2010: Started 
 *
 */

#ifndef _FREEFALL_H_
#define _FREEFALL_H_

// generic AVR includes:
#include<util/delay.h>
#include<avr/io.h>
#include<avr/interrupt.h>
#include<avr/wdt.h>
#include<inttypes.h>

// include all Freefall related libraries:
#include"adc.h"
#include"pwm.h"
#include"mirf.h"  // if the MIRF is unused, this code will be automatically optimized out
#ifndef NO_MMA
#include"mma7455.h"
#endif
#include"packetizer.h"
#include"spi.h"
#include"usart.h"
#include"wave.h"
#include"freefall_protocol.h"
//#include"usbdrv.h" // included by usbpack or usbmouse
//#include"usbpack.h" //user must include this because of conflict with usbmouse

#define LED0_PWM_CHANNEL 1
#define LED1_PWM_CHANNEL 0

#define CHG_CTRL	1 //portb.1
#define CHG_CTRL_PWM_CHANNEL 2 // timer1, output A = channel 2

#define PMON_GND	5 //portc.5
#define PMON 		4 //portc.4
#define BTEMP		3 //portc.3
#define CDS_PWR 	1 //portc.1
#define CDS		2 //portc.2

// chg_mode values:
#define SLOW		0
#define FAST		1

// Charging Temperatures
#define TEMP_FREEZING  	116*4 // 0C
#define TEMP_ROOM	163*4 // 20C
#define TEMP_CHG_START	186*4 // 30C = maximum charge start temperature
#define TEMP_CHG_FINISH 200*4 // 35C = charge shut-off temperature
#define TEMP_MAX	209*4 // 40C = maximum charge temperature

// GLOBAL MODE FLAGS:
//uint8_t mode = 1<<MODE_PLUGGED_IN | 1<<MODE_ENABLED;
uint8_t mode = 1<<MODE_ENABLED;
uint8_t charge_duty = 0;

// wdt disabler: (NOTE: If you use a bootloader, the _bootloader_ MUST do this
void wdt_init(void) __attribute__((naked)) __attribute__((section(".init3")));
void wdt_init(void)
{
    MCUSR = 0;
    wdt_disable();

    return;
}

void force_reset() {
	//cli(); // disable interrupts
	// enable watchdog in system reset mode
	WDTCSR |= (1<<WDE) | (1<<WDCE); //two step operation required 
	WDTCSR = 1<<WDE; //(step two)
	while(1); // wait for watchdog to reset processor
	// jump to bootloader:
	/*asm volatile("ldi r31,12 \n\t" 
		     "ldi r30,0 \n\t"
		     "ijmp"::);*/
}

uint8_t pack_float(float f, const float min, const float max) {
    if (f>max) f=max;
    if (f<min) f=min;
    return 255.*(f-min)/(max-min);
}

float unpack_float(uint8_t packed, float min, float max) {
    return packed/255.*(max-min)+min;
}


void cds_mode_pullup()
{
	DDRC |= 1 << CDS_PWR;		//make PC1 output
	PORTC |= (1<<CDS_PWR);
}

void cds_mode_pulldown()
{
	DDRC |= 1 << CDS_PWR;		//make PC1 output
	PORTC &= ~(1<<CDS_PWR);
}

void cds_mode_hz()
{
	DDRC &= ~(1 << CDS_PWR);	//make PC1 input (HZ)
	PORTC &= ~(1<<CDS_PWR); 	//disable AVRs internal pullup
}

void cds_init()
{
	DDRC |= 1 << CDS_PWR;		//make PC1 output
	DDRC &= ~(1 << CDS);		//make PC2 input 

	cds_mode_pullup();
}

uint8_t cds_read()
{
	uint8_t retval;

	adc_set_reference(ADC_REFERENCE_AVCC);

	//PORTC |= (1<<CDS_PWR);		//power voltage divider

	retval = adc_read(CDS);		//read PC2 (ignore the inaccurate reading)
	retval = adc_read(CDS);		//read PC2 
	
	//PORTC &= ~(1<<CDS_PWR); 	//shutdown to save power

	return 255-retval;
}

float cds_voltage()
{
    return adc_read_10bit(CDS)*3.3/(1<<10);
}

float cds_percent()
{
    return 100.-adc_read_10bit(CDS)*100./(1<<10);
}



void led0_duty(uint8_t duty) {
	pwm_set(LED0_PWM_CHANNEL, duty);
}

void led1_duty(uint8_t duty) {
	pwm_set(LED1_PWM_CHANNEL, duty);
}

uint8_t btemp_read()
{
	uint8_t retval;

	retval = adc_read(BTEMP);

	return retval;
}

void pman_slow() {
	// set charge control to slowest
    charge_duty = 0;
	pwm_set(CHG_CTRL_PWM_CHANNEL, charge_duty);
}

void pman_fast() {
	// set charge control to fastest
    charge_duty = 255;	
	pwm_set(CHG_CTRL_PWM_CHANNEL, charge_duty);
}

void pman_set_rate(uint8_t duty) {
    charge_duty = duty;
	pwm_set(CHG_CTRL_PWM_CHANNEL, charge_duty);
}

void pman_init() {
	DDRC &= ~(1<<PMON);
	DDRC &= ~(1<<BTEMP);
	DDRB |= (1<<CHG_CTRL);
    charge_duty = 0;
	pwm_init(CHG_CTRL_PWM_CHANNEL,8);
	pwm_set(CHG_CTRL_PWM_CHANNEL, charge_duty);
}

uint16_t pman_vext_10bit() {
    uint16_t retval;	
    adc_set_reference(ADC_REFERENCE_AVCC);  //3.3V ref
    DDRC |= 1<<PMON_GND;                    // power on voltdiv 
    PORTC &= ~(1<<PMON_GND);
    pwm_set(CHG_CTRL_PWM_CHANNEL, 255);     // set FAST charge
    adc_read_10bit(PMON);                   // ignore first result
    retval = adc_read_10bit(PMON);
    DDRC &= ~(1<<PMON_GND);                 // shutdown voltdiv
    pwm_set(CHG_CTRL_PWM_CHANNEL, charge_duty); // restore charge rate
    return retval;
}

uint16_t pman_vbatt_10bit() {
    uint16_t retval;	
    adc_set_reference(ADC_REFERENCE_AVCC);  //3.3V ref
    DDRC |= 1<<PMON_GND;                    // power on voltdiv 
    PORTC &= ~(1<<PMON_GND);
    pwm_set(CHG_CTRL_PWM_CHANNEL, 0);     // set SLOW charge
    adc_read_10bit(PMON);                   // ignore first result
    retval = adc_read_10bit(PMON);
    DDRC &= ~(1<<PMON_GND);                 // shutdown voltdiv
    pwm_set(CHG_CTRL_PWM_CHANNEL, charge_duty); // restore charge rate
    return retval;
}

#define pman_vusb pman_vext
float pman_vext() {
    // convert 10-bit adc reading to voltage
    return pman_vext_10bit()*3.3/(1<<10)*(1.5+6.98)/1.5;
}

float pman_vbatt() {
    // convert 10-bit adc reading to voltage
    return pman_vbatt_10bit()*3.3/(1<<10)*(1.5+6.98)/1.5;
}

uint16_t btemp_10bit() {
    adc_set_reference(ADC_REFERENCE_AVCC);
    return adc_read_10bit(BTEMP);
}


float btemp_voltage() {
    // convert 10-bit adc reading to voltage
    return btemp_10bit()*3.3/(1<<10);
}

float btemp_C() {
    //gives battery temperature in degrees C
    return (btemp_voltage()-.5)/.01;
}
void freefall_init() {

	wdt_enable(WDTO_4S); // anti-brick

	// set data direction TODO: this should be overridden by *_init()
	DDRB = 0xEF;
	DDRC = 0b00000100;
	DDRD = 0b11110100;

	adc_init();
	cds_init();
    pman_init();
	pwm_init(LED0_PWM_CHANNEL,8);
	pwm_init(LED1_PWM_CHANNEL,8);
	led0_duty(255); // turn on leds
	led1_duty(255);
#ifndef NO_MMA
	mma7455_init();
#endif
	mirf_init();
	_delay_ms(100);
#ifndef NO_MMA
	mma7455_config();    // if MMA does not respond, program will hang here
#endif
	led0_duty(0);   // turn off leds, meaning that MMA is functioning
	led1_duty(0);
	//usart_init(); // USB is the default comm method
	//usb_init(); // USB will be initialized by usbpack or usbmouse
}



#endif /* _FREEFALL_H_*/
