/*
 * mirf.h - driver for the nRF24L01 2.4GHz transciever module
 *
 
    Copyright (c) 2007 Stefan Engelke <mbox@stefanengelke.de>
    Copyright (c) 2010 Stingworks

    Permission is hereby granted, free of charge, to any person 
    obtaining a copy of this software and associated documentation 
    files (the "Software"), to deal in the Software without 
    restriction, including without limitation the rights to use, copy, 
    modify, merge, publish, distribute, sublicense, and/or sell copies 
    of the Software, and to permit persons to whom the Software is 
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be 
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
    DEALINGS IN THE SOFTWARE.

 * Todo:
 * 	On send, block until last TX is finished
 *
 * File History:
 *  
 *   11 FEB 10: imported from Stefan Engelke's mirf.h and mirf.c
 *   		updated defines for plus version
 *   12 FEB 10: completed port from Stefan's code
 *   		transmission working without interrupts
 *   16 FEB 10: cleaned up interface
 *
 */

#ifndef _MIRF_H_
#define _MIRF_H_

#include <avr/io.h>
#include "spi.h"

//*** mirf configuration ***
// These may be overridden in main.c before including this file.
#ifndef MIRF_CHANNEL
	#define MIRF_CHANNEL	2
#endif
#ifndef MIRF_PAYLOAD  // This payload must be the same on sender and reciever.
	#define MIRF_PAYLOAD	8 
#endif
#ifndef MIRF_CE // chip enabled line
	#define MIRF_CE		PD4
#endif
#ifndef MIRF_CSN // chip select line
	#define MIRF_CSN	PD7
#endif

#define MIRF_PORT PORTD
#define MIRF_DDR  DDRD




#define MIRF_CONFIG     ( (1<<MASK_RX_DR) | (1<<EN_CRC) | (0<<CRCO) )


// Macros
#define MIRF_CSN_HI     MIRF_PORT |=  (1<<MIRF_CSN);
#define MIRF_CSN_LO     MIRF_PORT &= ~(1<<MIRF_CSN);
#define MIRF_CE_HI      MIRF_PORT |=  (1<<MIRF_CE);
#define MIRF_CE_LO      MIRF_PORT &= ~(1<<MIRF_CE);
#define TX_POWERUP mirf_config_register(CONFIG, MIRF_CONFIG | ( (1<<PWR_UP) | (0<<PRIM_RX) ) )
#define RX_POWERUP mirf_config_register(CONFIG, MIRF_CONFIG | ( (1<<PWR_UP) | (1<<PRIM_RX) ) )



// Defines from nRF24L01+ datasheet:

/* Changes from nRF24L01 (non-plus version):
 * Register name CD -> RPD at 0x09
 * Register DYNPD added with bit memonics
 * Register FEATURE added with bit memonics
 *
 * LNA_HCURR bit obsoleted
 *
*/

/* Memory Map
 * Register addresses 18 to 1B are reserved for testing.
*/
#define CONFIG      0x00
#define EN_AA       0x01
#define EN_RXADDR   0x02
#define SETUP_AW    0x03
#define SETUP_RETR  0x04
#define RF_CH       0x05
#define RF_SETUP    0x06
#define STATUS      0x07
#define OBSERVE_TX  0x08
#define RPD_REG     0x09	//naming conflict with RPD bit
#define RX_ADDR_P0  0x0A
#define RX_ADDR_P1  0x0B
#define RX_ADDR_P2  0x0C
#define RX_ADDR_P3  0x0D
#define RX_ADDR_P4  0x0E
#define RX_ADDR_P5  0x0F
#define TX_ADDR     0x10
#define RX_PW_P0    0x11
#define RX_PW_P1    0x12
#define RX_PW_P2    0x13
#define RX_PW_P3    0x14
#define RX_PW_P4    0x15
#define RX_PW_P5    0x16
#define FIFO_STATUS 0x17
#define DYNPD       0x1C	//added
#define FEATURE     0x1D	//added


/* Bit Mnemonics */
#define MASK_RX_DR  6
#define MASK_TX_DS  5
#define MASK_MAX_RT 4
#define EN_CRC      3
#define CRCO        2
#define PWR_UP      1
#define PRIM_RX     0

#define ENAA_P5     5
#define ENAA_P4     4
#define ENAA_P3     3
#define ENAA_P2     2
#define ENAA_P1     1
#define ENAA_P0     0

#define ERX_P5      5
#define ERX_P4      4
#define ERX_P3      3
#define ERX_P2      2
#define ERX_P1      1
#define ERX_P0      0

#define AW          0

#define ARD         4
#define ARC         0

#define RF_CH_BIT    0		//added, conflict with RF_CH register

#define CONT_WAVE   7
#define RF_DR_LOW   5		//added
#define PLL_LOCK    4
#define RF_DR_HIGH  3		//changed from RF_DR
#define RF_PWR      1
#define RX_DR       6
#define TX_DS       5
#define MAX_RT      4
#define RX_P_NO     1
#define ST_TX_FULL  0		//naming conflict with FIFO_STATUS register
#define PLOS_CNT    4
#define ARC_CNT     0
#define RPD_BIT	    0		//naming conflict with RPD register
#define RX_PW_P0B   0
#define RX_PW_P1B   0
#define RX_PW_P2B   0
#define RX_PW_P3B   0
#define RX_PW_P4B   0
#define RX_PW_P5B   0

#define TX_REUSE    6
#define FST_TX_FULL 5		//naming conflict with STATUS register
#define TX_EMPTY    4
#define RX_FULL     1
#define RX_EMPTY    0

#define DPL_P5	    5
#define DPL_P4	    4
#define DPL_P3	    3
#define DPL_P2	    2
#define DPL_P1	    1
#define DPL_P0	    0

#define EN_DPL	    2
#define EN_ACK_PAY  1
#define EN_DYN_ACK  0


/* Instruction Mnemonics */
#define R_REGISTER    0x00
#define W_REGISTER    0x20
#define REGISTER_MASK 0x1F
#define R_RX_PAYLOAD  0x61
#define W_TX_PAYLOAD  0xA0
#define FLUSH_TX      0xE1
#define FLUSH_RX      0xE2
#define REUSE_TX_PL   0xE3
#define NOP           0xFF

uint8_t failed_packets;

void mirf_init() 
// Initializes pins to communicate with the MiRF module
// Should be called in the early initializing phase at startup.
{
    // Define MIRF_CSN and MIRF_CE as Output and set them to default
    MIRF_DDR |= ((1<<MIRF_CSN)|(1<<MIRF_CE));
    MIRF_CE_LO;
    MIRF_CSN_HI;

    // Initialize spi module
    spi_init();

    failed_packets = 0;
}


void mirf_config() 
// Sets the important registers in the MiRF module and powers the module
// in receiving mode
{
    // Set RF channel
    mirf_config_register(RF_CH,MIRF_CHANNEL);

    // Set RF channel
    // Set length of incoming payload 
    mirf_config_register(RX_PW_P0, MIRF_PAYLOAD);

    // Start receiver 
    RX_POWERUP;     // Power up in receiving mode

    MIRF_CE_HI;     // Listening for pakets

}

void mirf_set_RADDR(uint8_t * adr) 
// Sets the receiving address
{
    MIRF_CE_LO;
    mirf_write_register(RX_ADDR_P0,adr,5);
    MIRF_CE_HI;
}

void mirf_set_TADDR(uint8_t * adr)
// Sets the transmitting address
{
    mirf_write_register(TX_ADDR, adr,5);
}

extern uint8_t mirf_data_ready() 
// Checks if data is available for reading
{

    uint8_t status;

    // Read MiRF status 
    MIRF_CSN_LO;                                // Pull down chip select
    status = spi_transfer_byte(NOP);            // Read status register
    MIRF_CSN_HI;                                // Pull up chip select

    return status & (1<<RX_DR);
    //return 0;
}

extern void mirf_get_data(uint8_t * data) 
// Reads MIRF_PAYLOAD bytes into data array
{
    MIRF_CSN_LO;                               // Pull down chip select
    spi_transfer_byte( R_RX_PAYLOAD );         // Send cmd to read rx payload
    spi_transfer(data,data,MIRF_PAYLOAD);      // Read payload
    MIRF_CSN_HI;                               // Pull up chip select
    mirf_config_register(STATUS,(1<<RX_DR));   // Reset status register
}

void mirf_config_register(uint8_t reg, uint8_t value)
// Clocks only one byte into the given MiRF register
{
    MIRF_CSN_LO;
    spi_transfer_byte(W_REGISTER | (REGISTER_MASK & reg));
    spi_transfer_byte(value);
    MIRF_CSN_HI;
}

void mirf_read_register(uint8_t reg, uint8_t * value, uint8_t len)
// Reads an array of bytes from the given start position in the MiRF registers.
{
    MIRF_CSN_LO;
    spi_transfer_byte(R_REGISTER | (REGISTER_MASK & reg));
    spi_transfer(value,value,len);
    MIRF_CSN_HI;
}

void mirf_write_register(uint8_t reg, uint8_t * value, uint8_t len) 
// Writes an array of bytes into inte the MiRF registers.
{
    MIRF_CSN_LO;
    spi_transfer_byte(W_REGISTER | (REGISTER_MASK & reg));
    spi_write(value,len);
    MIRF_CSN_HI;
}


void mirf_send(uint8_t * value, uint8_t len) 
// Sends a data package to the default address. Be sure to send the correct
// amount of bytes as configured as payload on the receiver.
{

    //TODO: check to ensure that the previous packet was sent

    uint8_t status;


    MIRF_CE_LO;

    TX_POWERUP;                     // Power up

    MIRF_CSN_LO;                    // Pull down chip select
    status = spi_transfer_byte( FLUSH_TX );  // Write cmd to flush tx fifo
    MIRF_CSN_HI;                    // Pull up chip select

    if (status & 1<<MAX_RT) {       
	    // Last packet did not send, despite automatic resending
	    if (failed_packets != 255)  // Don't overflow counter
		    failed_packets++; 

	    // Enable further communication by clearing MAX_RT flag:
	    mirf_config_register(STATUS,1<<MAX_RT);
    } else {
	    failed_packets = 0;
    }

    MIRF_CSN_LO;                    // Pull down chip select
    spi_transfer_byte( W_TX_PAYLOAD ); // Write cmd to write payload
    spi_write(value,len);           // Write payload
    MIRF_CSN_HI;                    // Pull up chip select
    
    MIRF_CE_HI;                     // Start transmission
}



#endif /* _MIRF_H_ */
