/*
 * mma7455.h - driver for the MMA7455 SPI/I2C accelerometer
 *
 
    Copyright (c) 2010 Stingworks

    Permission is hereby granted, free of charge, to any person 
    obtaining a copy of this software and associated documentation 
    files (the "Software"), to deal in the Software without 
    restriction, including without limitation the rights to use, copy, 
    modify, merge, publish, distribute, sublicense, and/or sell copies 
    of the Software, and to permit persons to whom the Software is 
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be 
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
    DEALINGS IN THE SOFTWARE.

 *
 * File History:
 *  
 *  17 FEB 10: initial write using pseudo-code from manufacturer appnote
 *
 */

#ifndef _MMA7455_H__
#define _MMA7455_H_

#include <avr/io.h>
#include <avr/eeprom.h>
#include "spi.h"

// end-of-EEPROM calibration storage addresses:
const uint8_t * MMA7455_XOFF_EEPROM = 500;
const uint8_t * MMA7455_YOFF_EEPROM = 501;
const uint8_t * MMA7455_ZOFF_EEPROM = 502;

//*** mma7455 configuration ***
// These may be overridden in main.c before including this file.
#ifndef MMA7455_PORT // port of AVR to which the mma7455 module is connected
	#define MMA7455_PORT	PORTB
	#define MMA7455_DDR	DDRB
#endif
#ifndef MMA7455_CS // chip select 
	#define MMA7455_CS	PB0
#endif



// Defines from datasheet:

// registers
#define MMA7455_XOUTL	   0x00
#define MMA7455_XOUTH	   0x01
#define MMA7455_YOUTL	   0x02
#define MMA7455_YOUTH	   0x03
#define MMA7455_ZOUTL	   0x04
#define MMA7455_ZOUTH	   0x05
#define MMA7455_XOUT8	   0x06
#define MMA7455_YOUT8	   0x07
#define MMA7455_ZOUT8	   0x08
#define MMA7455_STATUS     0x09
#define MMA7455_DETSRC     0x0A
#define MMA7455_TOUT       0x0B
#define MMA7455_RESERVED0  0x0C
#define MMA7455_I2CAD	   0x0D
#define MMA7455_USRINF     0x0E
#define MMA7455_WHOAMI     0x0F
#define MMA7455_XOFFL      0x10
#define MMA7455_XOFFH	   0x11
#define MMA7455_YOFFL	   0x12
#define MMA7455_YOFFH      0x13
#define MMA7455_ZOFFL      0x14
#define MMA7455_ZOFFH      0x15
#define MMA7455_MCTL       0x16
#define MMA7455_INTRST     0x17
#define MMA7455_CTL1       0x18
#define MMA7455_CTL2       0x19
#define MMA7455_LDTH	   0x1A
#define MMA7455_PDTH       0x1B
#define MMA7455_PW         0x1C
#define MMA7455_LT         0x1D
#define MMA7455_TW         0x1E
#define MMA7455_RESERVED1  0x1F

//bits
//for Status Register (STATUS)
//D7:D3 unused
#define MMA7455_PERR		2
#define MMA7455_DOVR		1
#define MMA7455_DRDY		0

//for Detection Source Register (DETSCR)
#define MMA7455_LDX        7
#define MMA7455_LDY        6
#define MMA7455_LDZ        5
#define MMA7455_PDX        4
#define MMA7455_PDY        3
#define MMA7455_PDZ        2
#define MMA7455_INT1       1
#define MMA7455_INT2       0

//for I2C Device Address Register (I2CAD)
#define MMA7455_I2CDIS		7


//for Mode Control Register (MCTL)
//#define MMA7455_           7
#define MMA7455_DRPD       6
#define MMA7455_SPI3W      5
#define MMA7455_STON       4
#define MMA7455_GLVL1      3
#define MMA7455_GLVL0      2
#define MMA7455_MODE1      1
#define MMA7455_MODE0      0

//for Interrupt Latch Reset (INTRST)
#define MMA7455_CLR_INT2   1
#define MMA7455_CLR_INT1   0

//for Control 1 (CTL1)
#define MMA7455_DFBW       7
#define MMA7455_THOPT      6
#define MMA7455_ZDA        5
#define MMA7455_YDA        4
#define MMA7455_XDA        3
#define MMA7455_INTREG1    2
#define MMA7455_INTREG0    1
#define MMA7455_INTPIN     0

//for Control 2 (CTL2)
#define MMA7455_DRVO       2
#define MMA7455_PDPL       1
#define MMA7455_LDPL       0




// Macros
#define MMA7455_CS_HI      MMA7455_PORT |=  (1<<MMA7455_CS);
#define MMA7455_CS_LO      MMA7455_PORT &= ~(1<<MMA7455_CS);

#define MMA7455_CALIBRATION_SAMPLES 16  // range: 1-127

// Change the type of a (u)int8_t variable
//  without changing the underlying data.
//  This is a "change of interpretation".
//  These functions are needed because otherwise
//  a type cast would occur, causing an unwanted
//  change in the underlying data.
int8_t uint8_to_int8(uint8_t u) {
	return * ((int8_t *) &u);
}
uint8_t int8_to_uint8(int8_t s) {
	return * ((uint8_t *) &s);
}



int8_t mma7455_xoff, mma7455_yoff, mma7455_zoff;

void mma7455_load_calibration()
{
    //load calib data from EEPROM
    mma7455_xoff = uint8_to_int8(eeprom_read_byte(MMA7455_XOFF_EEPROM));
    mma7455_yoff = uint8_to_int8(eeprom_read_byte(MMA7455_YOFF_EEPROM));
    mma7455_zoff = uint8_to_int8(eeprom_read_byte(MMA7455_ZOFF_EEPROM));
}

void mma7455_save_calibration()
{
    //save calib data to EEPROM
    eeprom_update_byte(MMA7455_XOFF_EEPROM,int8_to_uint8(mma7455_xoff));
    eeprom_update_byte(MMA7455_YOFF_EEPROM,int8_to_uint8(mma7455_yoff));
    eeprom_update_byte(MMA7455_ZOFF_EEPROM,int8_to_uint8(mma7455_zoff));
}



uint8_t mma7455_get_reg(uint8_t reg)
// Retrieve the value of an MMA7455 register
{
	uint8_t value;
	MMA7455_CS_LO;
	spi_transfer_byte((reg & 0x3F) << 1);  // select reg
	value = spi_transfer_byte(0x00);      // read value
	MMA7455_CS_HI;
	return value;
}

void mma7455_set_reg(uint8_t reg, uint8_t value)
// Clocks only one byte into the given MMA7455 register
{
	MMA7455_CS_LO;
	spi_transfer_byte(((reg & 0x3F) << 1) | 0x80); // select reg
	spi_transfer_byte(value);
	MMA7455_CS_HI;
}


#ifdef MMA7455_HW_CALIBRATION
int8_t mma7455_read_x() {
	return uint8_to_int8(mma7455_get_reg(MMA7455_XOUT8));
}
int8_t mma7455_read_y() {
	return uint8_to_int8(mma7455_get_reg(MMA7455_YOUT8));
}
int8_t mma7455_read_z() {
	return uint8_to_int8(mma7455_get_reg(MMA7455_ZOUT8));
}
#else //soft calib
int8_t mma7455_read_x() {
	return uint8_to_int8(mma7455_get_reg(MMA7455_XOUT8))+mma7455_xoff;
}
int8_t mma7455_read_y() {
	return uint8_to_int8(mma7455_get_reg(MMA7455_YOUT8))+mma7455_yoff;
}
int8_t mma7455_read_z() {
	return uint8_to_int8(mma7455_get_reg(MMA7455_ZOUT8))+mma7455_zoff;
}
#endif






void mma7455_init() 
// Initializes pins to communicate with the MMA7455 module
// Should be called in the early initializing phase at startup.
{
    MMA7455_DDR |= (1<<MMA7455_CS); // Slave-select is output pin

    MMA7455_CS_HI;

    // Initialize spi module
    spi_init();

    mma7455_xoff = mma7455_yoff = mma7455_zoff = 0;
}


void mma7455_config()
// Sets the important registers in the MMA7455 module
{
	// i2c address reg
	mma7455_set_reg(MMA7455_I2CAD,
			1<<MMA7455_I2CDIS);	// disable i2c

	// mode control reg
	mma7455_set_reg(MMA7455_MCTL,
			0<<MMA7455_STON | // self-test mode is OFF
			0<<MMA7455_MODE1 |	// measurement mode MODE1:0 = 01
			1<<MMA7455_MODE0 |	// 
			1<<MMA7455_GLVL1 |	// set to 4g mode, GLVL1:0
			0<<MMA7455_GLVL0);	// 00 8g, 10 4g, 01 2g

	// self-test is not enabled
	// spi is in 4-wire mode
	// data ready status is output to INT1 pin
	

	// control 1 reg
	mma7455_set_reg(MMA7455_CTL1,0);
			//1<<MMA7455_DFBW);	// set digital filter
			// 62.5Hz filter (0) => 125Hz smpl freq (8ms)
			// 125Hz filter (1) => 250Hz sampl freq (4ms)

    mma7455_load_calibration();
}

uint8_t mma7455_data_ready()
// Check status register for new data (data ready flag)
{
	uint8_t status;
	status = mma7455_get_reg(MMA7455_STATUS);

	if((status & 0x01) == 1)
		return 1;

	return 0;
}

void mma7455_calibrate() {
    // WARNING! This assumes that the board is in the
    //  standard orientation! (z = +1g, 4g mode)
	mma7455_calibrate_custom(0,0,32);
    mma7455_save_calibration();
    //mma7455_config();
}

void mma7455_calibrate_custom(int8_t xs, int8_t ys, int8_t zs)
// Calibration procedure as per manu app note AN3745
// Assumes that the accelerometer is laying flat
//  i.e. with x=0 y=0 z=+1g.
//
//  NOTE: this is only valid for 4g 8-bit mode.
//  TODO: allow any gmode to be used.
//
//  It then takes several samples and sets the offset regs accordingly.
{
	int16_t x = 0, y = 0, z = 0;
	uint8_t i;
    mma7455_xoff = mma7455_yoff = mma7455_zoff = 0;
	for (i=0;i<MMA7455_CALIBRATION_SAMPLES;i++) {
		//while(mma7455_data_ready() != 1);
        _delay_ms(4);
		x += mma7455_read_x();
		y += mma7455_read_y();
		z += mma7455_read_z();
	}
	x /= MMA7455_CALIBRATION_SAMPLES; 
	y /= MMA7455_CALIBRATION_SAMPLES;
	z /= MMA7455_CALIBRATION_SAMPLES;

	x -= xs;
	y -= ys;
	z -= zs;

#ifdef MMA7455_HW_CALIBRATION
	// NOTE: The HW calibration sets the correct calib vals,
	//  but it does not seem to have the right effect on
	//  the resulting readings... (?!?!?)

	x *= -2; // offset drift regs are in 1/2 lsb units
	y *= -2;
	z *= -2;

	mma7455_set_reg(MMA7455_XOFFL, int8_to_uint8(x));
	mma7455_set_reg(MMA7455_YOFFL, int8_to_uint8(y));
	mma7455_set_reg(MMA7455_ZOFFL, int8_to_uint8(z));

#else // software calibration by default
	mma7455_xoff = -x;
	mma7455_yoff = -y;
	mma7455_zoff = -z;
#endif

}

uint8_t mma7455_data_overrun()
//Check status register for missed data (data overwritten flag)
{
	uint8_t status;
	status = mma7455_get_reg(MMA7455_STATUS);

	if((status && 0x02) == 1)
		return 1;

	return 0;
}

#endif /* _MMA7455_H_ */
