/*
 * packetizer.h - a simple packetizing protocol
 *
 *  
 *  Copyright (c) 2010 Stingworks
 *
 
    Permission is hereby granted, free of charge, to any person 
    obtaining a copy of this software and associated documentation 
    files (the "Software"), to deal in the Software without 
    restriction, including without limitation the rights to use, copy, 
    modify, merge, publish, distribute, sublicense, and/or sell copies 
    of the Software, and to permit persons to whom the Software is 
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be 
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
    DEALINGS IN THE SOFTWARE.

 * NOTE: This code may be run on an AVR and/or on a PC.
 *
 * File History:
 *  
 *  16 FEB 2010:  begun
 *
 */

#include<inttypes.h>

// you MAY change the PACKETIZER_ESCAPE value:
#define PACKETIZER_ESCAPE 		0x5C

// you MAY NOT change these values:
#define PACKETIZER_ESCAPE_ESCAPE 	0xFE

#define PACKETIZER_ERROR_CHECKSUM	0xFF
#define PACKETIZER_ERROR_OVERRUN	0xFD

// note: if the salt is set to 0x00, then packets consisting
//  entirely of zeros will be recognized as sent correctly,
//  which should be avoided
#define PACKETIZER_CHECKSUM_SALT	0x07

// function pointers to external io
uint8_t (*packetizer_getc)(void) = 0;
void (*packetizer_putc)(uint8_t) = 0;

void packetizer_init(uint8_t (*getc_func) (void), void (*putc_func) (uint8_t))
// packetizer initialization
//  you must call this function once to supply the external io functions
//  if you only need one of getc and putc, pass null for the unneeded pointer
{
	packetizer_getc = getc_func;
	packetizer_putc = putc_func;
}

void packetizer_write(uint8_t * data, uint8_t length)
// send a packet using a simple protocol
//  0 =< length =< 127
{
	uint8_t i;
	uint8_t checksum = PACKETIZER_CHECKSUM_SALT;

	if (packetizer_putc == 0) return;

	// send packet header
	packetizer_putc(PACKETIZER_ESCAPE);
	packetizer_putc(length);

	// send packet payload, escaping out any escape charicters
	for (i = 0; i<length; i++) {
		packetizer_putc(data[i]);
		if (data[i] == PACKETIZER_ESCAPE)
			packetizer_putc(PACKETIZER_ESCAPE_ESCAPE);
		checksum += data[i];
	}

	packetizer_putc(checksum);
	if (checksum == PACKETIZER_ESCAPE)
		packetizer_putc(PACKETIZER_ESCAPE_ESCAPE);


}

uint8_t packetizer_read(uint8_t * data, uint8_t max_length)
// reads the next packet sent into data array
// WARNING! This is a BLOCKING read
//  number of bytes actually read is returned
//  a return value of zero indicates that an error occured
{
	if (packetizer_getc == 0) 
		return 0; // you need to initialize 

	uint8_t length;
	uint8_t c;
	uint8_t i;
	uint8_t checksum = PACKETIZER_CHECKSUM_SALT;
	while (1) {
		if (packetizer_getc() == PACKETIZER_ESCAPE) {
			c = packetizer_getc();
		       	if (c != PACKETIZER_ESCAPE_ESCAPE) {
				length = c;
				if (length>max_length)
					return PACKETIZER_ERROR_OVERRUN;
				for (i=0;i<length;i++) {
					c = packetizer_getc();
					if (c == PACKETIZER_ESCAPE)
						packetizer_getc();
					data[i] = c;
					checksum += c;
				}
				c = packetizer_getc();

				if (c != checksum)
					return PACKETIZER_ERROR_CHECKSUM;
				return length;
			}
		}
	}
}

