/*
 * usbpack.h - sends configurable-length packets over USB
 *
 *   Copyright (c) 2010 Stingworks
 *
 * File History:
 *  
 *   21 July 2010:
 *
 *
 * Original Copyright Notice:
 *   
 *   Based on code by Christian Starkjohann for interfacing
 *    with Objective Development's VUSB library. See usbdrv
 *    directory for details.
 *
 */

#ifndef _USBPACK_H_
#define _USBPACK_H_

#include "usbdrv.h"

uint8_t packet_out[FREEFALL_PACKET_LENGTH];	// note: set packet length in "freefall_protocol.h"
uint8_t packet_in[FREEFALL_PACKET_LENGTH];	// note: set packet length in "freefall_protocol.h"
uint8_t packets_sent; 		//rolling variable, may be cleared by user program
uint8_t packets_received;	//rolling variable, may be cleared by user program

/* ------------------------------------------------------------------------- */
/* ----------------------------- USB interface ----------------------------- */
/* ------------------------------------------------------------------------- */

PROGMEM char usbHidReportDescriptor[22] = {	/* USB report descriptor */
	0x06, 0x00, 0xff,	// USAGE_PAGE (Generic Desktop)
	0x09, 0x01,		// USAGE (Vendor Usage 1)
	0xa1, 0x01,		// COLLECTION (Application)
	0x15, 0x00,		//   LOGICAL_MINIMUM (0)
	0x26, 0xff, 0x00,	//   LOGICAL_MAXIMUM (255)
	0x75, 0x08,		//   REPORT_SIZE (8)
	0x95, 0x80,		//   REPORT_COUNT (128)
	0x09, 0x00,		//   USAGE (Undefined)
	0xb2, 0x02, 0x01,	//   FEATURE (Data,Var,Abs,Buf)
	0xc0			// END_COLLECTION
};

/* Since we define only one feature report, we don't use report-IDs (which
 * would be the first byte of the report). The entire report consists of 128
 * opaque data bytes.
 */

/* The following variables store the status of the current data transfer */
static uchar currentAddress;
static uchar bytesRemaining;

/* ------------------------------------------------------------------------- */

/* usbFunctionRead() is called when the host requests a chunk of data from
 * the device. For more information see the documentation in usbdrv/usbdrv.h.
 */
uchar usbFunctionRead(uchar * data, uchar len)
{
	if (len > bytesRemaining)
		len = bytesRemaining;
	memcpy(data, (uchar *) packet_out + currentAddress, len);
	currentAddress += len;
	bytesRemaining -= len;
	return len;
}

/* usbFunctionWrite() is called when the host sends a chunk of data to the
 * device. For more information see the documentation in usbdrv/usbdrv.h.
 */
uchar usbFunctionWrite(uchar * data, uchar len)
{
	if (bytesRemaining == 0) {
		return 1;	/* end of transfer */
	}
	if (len > bytesRemaining) {
		len = bytesRemaining;
	}
	memcpy((uchar *) packet_in + currentAddress, data, len);
	currentAddress += len;
	bytesRemaining -= len;
	return bytesRemaining == 0;	/* return 1 if this was the last chunk */
}

/* ------------------------------------------------------------------------- */

usbMsgLen_t usbFunctionSetup(uchar data[8])
{
	usbRequest_t *rq = (void *)data;

	if ((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS) {	/* HID class request */
		if (rq->bRequest == USBRQ_HID_GET_REPORT) {	/* wValue: ReportType (highbyte), ReportID (lowbyte) */
			/* since we have only one report type, we can ignore the report-ID */
			bytesRemaining = 8;
			currentAddress = 0;
			packets_sent++;
			return USB_NO_MSG;	/* use usbFunctionRead() to obtain data */
		} else if (rq->bRequest == USBRQ_HID_SET_REPORT) {
			/* since we have only one report type, we can ignore the report-ID */
			bytesRemaining = 8;
			currentAddress = 0;
			packets_received++;
			return USB_NO_MSG;	/* use usbFunctionWrite() to receive data from host */
		}
	} else {
		/* ignore vendor type requests, we don't use any */
	}
	return 0;
}

/* ------------------------------------------------------------------------- */

/* Serial Number control is managed by USBPACK: */

const uint8_t * FREEFALL_SERIAL_EEPROM_ADDR = 507;
#define SERIAL_NUMBER_LENGTH 5
static int serialNumberDescriptor[SERIAL_NUMBER_LENGTH + 1];
uint8_t usbFunctionDescriptor(usbRequest_t * rq)
{
    uint8_t len = 0;
    usbMsgPtr = 0;
    if (rq->wValue.bytes[1] == USBDESCR_STRING && rq->wValue.bytes[0] == 3) {
        usbMsgPtr = (uint8_t *)serialNumberDescriptor;
        len = sizeof(serialNumberDescriptor);
    }
    return len;
}
void freefall_save_serial() {
    uint8_t i;
    uint8_t tmp;
    for (i=0; i<SERIAL_NUMBER_LENGTH; i++) {
        tmp = serialNumberDescriptor[i+1];
        eeprom_update_byte(FREEFALL_SERIAL_EEPROM_ADDR+i,tmp);
    }
}
void freefall_load_serial() {
    uint8_t tmp;
    uint8_t i;
    for (i=0; i<SERIAL_NUMBER_LENGTH; i++) {
        tmp = eeprom_read_byte(FREEFALL_SERIAL_EEPROM_ADDR+i);
        serialNumberDescriptor[i+1] = tmp; // type cast
    }
    serialNumberDescriptor[0] = 
        USB_STRING_DESCRIPTOR_HEADER(SERIAL_NUMBER_LENGTH);
}

/* -------------------------------------------------------------- */

void usbpack_init()
{
	packets_sent = 0;
	packets_received = 0;

    // save a new serial number if defined
    //  FFSN* are defined in a makefile
#ifdef FFSN1 
    serialNumberDescriptor[1] = FFSN10000;
    serialNumberDescriptor[2] = FFSN1000;
    serialNumberDescriptor[3] = FFSN100;
    serialNumberDescriptor[4] = FFSN10;
    serialNumberDescriptor[5] = FFSN1;
    freefall_save_serial();
#endif

    freefall_load_serial();
	usbInit();
	usbDeviceDisconnect();	/* enforce re-enumeration, do this while interrupts are disabled! */
	uint8_t i = 0;
	while (--i) {		/* fake USB disconnect for > 250 ms */
		wdt_reset();
		_delay_ms(1);
	}
	usbDeviceConnect();
	sei();
}

//usbPoll must be called often in user program
#define usbpack_update usbPoll

#endif				/* _USBPACK_H_ */
