/*
 * pwm.h - Pulse Width Modulation library
 *
 *  
 *  Copyright (c) 2000, 2003 Pascal Stang 
 *  Copyright (c) 2010 Stingworks
 *
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 *
 *
 * File History:
 *  
 *   10 FEB 2010:  ported from Procyon
 *
 */


#define TIMER_CLK_STOP		0x00	///< Timer Stopped
#define TIMER_CLK_DIV1		0x01	///< Timer clocked at F_CPU
#define TIMER_CLK_DIV8		0x02	///< Timer clocked at F_CPU/8
#define TIMER_CLK_DIV64		0x03	///< Timer clocked at F_CPU/64
#define TIMER_CLK_DIV256	0x04	///< Timer clocked at F_CPU/256
#define TIMER_CLK_DIV1024	0x05	///< Timer clocked at F_CPU/1024
#define TIMER_CLK_T_FALL	0x06	///< Timer clocked at T falling edge
#define TIMER_CLK_T_RISE	0x07	///< Timer clocked at T rising edge
#define TIMER_PRESCALE_MASK	0x07	///< Timer Prescaler Bit-Mask

void pwm_set_prescaler(unsigned char ch, unsigned char prescale) {

	// control the prescale of the associated timer,
	//  thus controlling the PWM period
	if (ch==0 || ch==1) {
		TCCR0B = TCCR0B & ~TIMER_PRESCALE_MASK | prescale;
	} else if (ch == 2 || ch == 3) {
		TCCR1B = TCCR1B & ~TIMER_PRESCALE_MASK | prescale;
	}
}

void pwm_init(unsigned char ch, unsigned char resolution) {

	if (ch==0 || ch==1) {
	
		// enable timer1 as 8bit phase-correct PWM
		//  (because other resolutions are not available
		//   on this timer)
		TCCR0A &= ~(1<<WGM01);
		TCCR0A |= 1<<WGM00;
	
		// clear output compare value A
		OCR0A = 0;
		// clear output compare value B
		OCR0B = 0;

	} else if (ch == 2 || ch == 3) {

		// enable timer1 as 8,9,10bit phase-correct PWM
		if(resolution == 9)
		{	// 9bit mode
			TCCR1A |= 1<<WGM11;
			TCCR1A &= ~(1<<WGM10);
		}
		else if(resolution == 10)
		{	// 10bit mode
			TCCR1A |= 1<<WGM11;
			TCCR1A |= 1<<WGM10;
		}
		else
		{	// default 8bit mode
			TCCR1A &= ~(1<<WGM11);
			TCCR1A |= 1<<WGM10;
		}
	
		// clear output compare value A
		OCR1AH = 0;
		OCR1AL = 0;
		// clear output compare value B
		OCR1BH = 0;
		OCR1BL = 0;
	}

	pwm_set_prescaler(ch,TIMER_CLK_DIV1);
	pwm_on(ch); // actually turn on the initialized channel
}



void pwm_shutdown(unsigned char ch) {
	
	// take the corresponding timer out of PWM mode
	//  NOTE: use pwm_off to shut off a single channel
	if (ch == 0 || ch == 1) {
	
		// turn off timer0 PWM mode
		TCCR0A &= ~(1<<WGM01);
		TCCR0A &= ~(1<<WGM00);
	
		// set PWM1A/B (OutputCompare action) to none
		pwm_off(0);
		pwm_off(1);
	} else {

		// turn off timer1 PWM mode
		TCCR1A &= ~(1<<WGM11);
		TCCR1A &= ~(1<<WGM10);
	
		// set PWM1A/B (OutputCompare action) to none
		pwm_off(2);
		pwm_off(3);
	}
}


void pwm_off(unsigned char ch) {

	// turn off a single channel 
	//  NOTE: if you want to shutdown the whole timer use pwm_shutdown instead
	if (ch == 0) {
		TCCR0A &= ~(1<<COM0A1);
		TCCR0A &= ~(1<<COM0A0);
	} else if (ch == 1) {
		TCCR0A &= ~(1<<COM0B1);
		TCCR0A &= ~(1<<COM0B0);
	} else if (ch == 2) {
		TCCR1A &= ~(1<<COM1A1);
		TCCR1A &= ~(1<<COM1A0);
	} else if (ch == 3) {
		TCCR1A &= ~(1<<COM1B1);
		TCCR1A &= ~(1<<COM1B0);
	}
}
		
void pwm_on(unsigned char ch) {

	// turn on a channel for non-inverted PWM output
	if (ch == 0) {
		TCCR0A |= 1<<COM0A1;
		TCCR0A &= ~(1<<COM0A0);
	} else if (ch == 1) {
		TCCR0A |= 1<<COM0B1;
		TCCR0A &= ~(1<<COM0B0);
	} else if (ch == 2) {
		TCCR1A |= 1<<COM1A1;
		TCCR1A &= ~(1<<COM1A0);
	} else if (ch == 3) {
		TCCR1A |= 1<<COM1B1;
		TCCR1A &= ~(1<<COM1B0);
	}
}

void pwm_invert(unsigned char ch) {

	// turn on a channel for inverted PWM output
	if (ch == 0) {
		TCCR0A |= 1<<COM0A1;
		TCCR0A |= 1<<COM0A0;
	} else if (ch == 1) {
		TCCR0A |= 1<<COM0B1;
		TCCR0A |= 1<<COM0B0;
	} else if (ch == 2) {
		TCCR1A |= 1<<COM1A1;
		TCCR1A |= 1<<COM1A0;
	} else if (ch == 3) {
		TCCR1A |= 1<<COM1B1;
		TCCR1A |= 1<<COM1B0;
	}
}

void pwm_set(unsigned char ch, unsigned int duty_cycle) {
	
	// set PWM (output compare) duty for specified channel
	// NOTE:	duty_cycle should be in the range 0-255 for 8bit PWM
	//		duty_cycle should be in the range 0-511 for 9bit PWM
	//		duty_cycle should be in the range 0-1023 for 10bit PWM
	if (ch == 0) {
		OCR0A = duty_cycle;
	} else if (ch == 1) {
		OCR0B = duty_cycle;
	} else if (ch == 2) {
		OCR1A = duty_cycle;
	} else if (ch == 3) {
		OCR1B = duty_cycle;
	}
}


